package Geo::Coder::Free::DB::openaddresses;

use strict;
use warnings;
use Database::Abstraction;

=head1 NAME

Geo::Coder::Free::DB::openaddresses - interface to data from http://results.openaddresses.io/

=head1 VERSION

Version 0.39

=cut

our $VERSION = '0.39';

our @ISA = ('Database::Abstraction');

1;
