package Geo::Coder::Free::Display::index;

use strict;
use warnings;

=head1 VERSION

Version 0.39

=cut

our $VERSION = '0.39';

# Display the index page

use Geo::Coder::Free::Display;

our @ISA = ('Geo::Coder::Free::Display');

1;
